/*
   This is a sample C program that shows how an application can write a
   congruent special value dataset. This program will write one dataset,
   "SnowDepth", that contains actual data.
   
   This sample code is similar to congruent_write.c, but different in that this
   code uses n-bit filter.

   This file assumes there are only two special values, and creates a
   customized datatype that can contain only two bits. Then, n-bit filter is
   applied to save the storage.

   To check if the n-bit filter actually reduces the file size, one can change
   the size of dimension from {10,7} to {1000,7}.

   For other information, please refer to comments in congruent_write.c.
 */

#define H5_USE_16_API
#include "hdf5.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NUM_SPECIALVALUES  2

#define errorexit()  _errorexit(__LINE__)

static void _errorexit(int line)
{
	fprintf(stderr, "errorexit at %d\n", line);
	exit(1);
}

static void create_file(const char *fname)
{
	hid_t fileid, spaceid, dsetid;
	const char *attrdesc;
	hsize_t dimsize[2] = { 10, 7 };
	int i, j, k;

	/* create an HDF5 file */
	if ((fileid = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT)) < 0) errorexit();
	/* create a dataspace for a dataset and a congruent dataset */
	if ((spaceid = H5Screate_simple(2, dimsize, NULL)) < 0) errorexit();

	/* define a dataset */
	{
		float *data;

		/* create an HDF5 dataset that represents actual data */
		if ((dsetid = H5Dcreate(fileid, "SnowDepth", H5T_NATIVE_FLOAT, spaceid, H5P_DEFAULT)) < 0) errorexit();

		/* fill actual data (some of them are garbages)
		        0  1  2  3  4  5  6 ...
		     +----------------------
		   0 |  1  2  3  4  5  6  7
		         ...
		*/
		if ((data = malloc(sizeof(data[0]) * dimsize[0] * dimsize[1])) == NULL) errorexit();
		for (k = i = 0; i < dimsize[0]; ++i) {
			for (j = 0; j < dimsize[1]; ++j)
				data[i * dimsize[1] + j] = ++k;
		}

		/* write actual data to the dataset we just created */
		if (H5Dwrite(dsetid, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) errorexit();

		free(data);

		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* define a congruent dataset */
	{
		hid_t congruenttid, prop, attrspace, attrtype, attrid;
		unsigned char *data;

		/* assume that special values can be hold in one byte */
		if (NUM_SPECIALVALUES > 8) errorexit();

		/* create a customized datatype */
		if ((congruenttid = H5Tcopy(H5T_NATIVE_UINT)) < 0) errorexit();
		if (H5Tset_precision(congruenttid, NUM_SPECIALVALUES) < 0) errorexit();
		if (H5Tset_offset(congruenttid, 8 - NUM_SPECIALVALUES) < 0) errorexit();

		/* apply the n-bit filter to save the storage */
		if ((prop = H5Pcreate(H5P_DATASET_CREATE)) < 0) errorexit();
		if (H5Pset_chunk(prop, 2, dimsize) < 0) errorexit();
		if (H5Pset_nbit(prop) < 0) errorexit();

		/* create an HDF5 dataset that represents special values for the dataset created above */
		if ((dsetid = H5Dcreate(fileid, "SpecialValues", congruenttid, spaceid, prop)) < 0) errorexit();

		/* fill non-mutual special values (0: no special values, 1: WaterFilled, 2: IceCovered, 3: both)
		        0  1  2  3  4  5  6
		     +----------------------
		   0 |  3  0  0  1  0  2  1
		   1 |  3  0  0  1  0  2  1
		   2 |  3  0  0  1  0  2  1
		   3 |  3  0  0  1  0  2  1
		   4 |  3  0  0  1  0  2  1
		*/
		if ((data = malloc(sizeof(data[0]) * dimsize[0] * dimsize[1])) == NULL) errorexit();
		for (i = 0; i < dimsize[0]; ++i) {
			for (j = 0; j < dimsize[1]; ++j) {
				unsigned char masks = 0;
				/* for each row, special value '1' is marked once every three elements */
				if (j % 3 == 0) masks |= 1;
				/* for each row, special value '2' is marked once every five elements */
				if (j % 5 == 0) masks |= 2;
				data[i * dimsize[1] + j] = masks;
			}
		}

		/* write special values to the congruent dataset */
		if (H5Dwrite(dsetid, H5T_NATIVE_UCHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) errorexit();

		/* (optional) write an attribute that describes this dataset, this attribute just helps the user know what the special value represents for. */
		attrdesc = "1: water-filled, 2: ice-covered";
		if ((attrspace = H5Screate(H5S_SCALAR)) < 0) errorexit();
		if ((attrtype = H5Tcopy(H5T_C_S1)) < 0) errorexit();
		if (H5Tset_size(attrtype, strlen(attrdesc)) < 0) errorexit();

		if ((attrid = H5Acreate(dsetid, "Description", attrtype, attrspace, H5P_DEFAULT)) < 0) errorexit();
		if (H5Awrite(attrid, attrtype, attrdesc) < 0) errorexit();
		if (H5Aclose(attrid) < 0) errorexit();

		free(data);

		/* close the customized datatype */
		if (H5Tclose(congruenttid) < 0) errorexit();

		/* close the dataset creation property */
		if (H5Pclose(prop) < 0) errorexit();

		/* close the dataset */
		if (H5Dclose(dsetid) < 0) errorexit();
	}

	/* close the dataspace */
	if (H5Sclose(spaceid) < 0) errorexit();
	/* close the file */
	if (H5Fclose(fileid) < 0) errorexit();
}

int main(int argc, char **argv)
{
	create_file("nbit.c.h5");
	return 0;
}

/* vi:set ts=4 sw=4 sts=4: */
