/*
   This is a sample C program that shows how an application can write attribute
   triplets that include special value information. This program will write
   one dataset, "SnowDepth", that contains actual data. Values of some dataset
   elements should be replaced by special value.

   This program will attach an attribute called "SpecialValues" to the
   "SnowDepth" dataset. This attribute is an array of triplets. A triplet
   consists of a description, a value and a dataset region. The dataset region
   contains the selection of elements in the dataset that a special value
   applies to. One can manipulate the dataset region using HDF5 APIs.

   For simplicity, we will use a small 5-by-7 array, and "SnowDepth" will be
   filled by meaningless numbers. We will use two special values. As the
   "description" field shows, the first special value is for "WaterFilled" and
   the second special value is for "IceCovered". We will not assume that two
   special values are mutually exclusive; in other words, both special values
   can be set for an element.

   For more information, please refer to the following sections from RFC:
   Special Values in HDF5 http://www.hdfgroup.org/pubs/rfcs/RFC_Special_Values_in_HDF5.pdf
      Section 4.3.4 "Attribute Triplet"
      Section 4.2.2 "Non-Mutually Exclusive Special Values"
 */

#define H5_USE_16_API
#include "hdf5.h"
#include <stdio.h>
#include <stdlib.h>

#define DATASET_RANK  2
#define NUM_SPECIALVALUES  2

typedef struct {
	const char *description;
	int value;
	/* the following type corresponds to H5R_DATASET_REGION */
	hdset_reg_ref_t region;
} triplet_t;

#define errorexit()  _errorexit(__LINE__)

static void _errorexit(int line)
{
	fprintf(stderr, "errorexit at %d\n", line);
	exit(1);
}

static void create_file(const char *fname)
{
	hid_t fileid, spaceid, dsetid;
	hsize_t dimsize[DATASET_RANK] = { 5, 7 };
	int i, j, k;

	/* create an HDF5 file */
	if ((fileid = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT)) < 0) errorexit();

	/* create a dataspace for a dataset */
	if ((spaceid = H5Screate_simple(DATASET_RANK, dimsize, NULL)) < 0) errorexit();

	/* create an HDF5 dataset that represents actual data */
	if ((dsetid = H5Dcreate(fileid, "SnowDepth", H5T_NATIVE_FLOAT, spaceid, H5P_DEFAULT)) < 0) errorexit();

	/* fill actual data (some of them are garbages)
		        0  1  2  3  4  5  6
		     +----------------------
		   0 |  1  2  3  4  5  6  7
		   1 |  8  9 10 11 12 13 14
		   2 | 15 16 17 18 19 20 21
		   3 | 22 23 24 25 26 27 28
		   4 | 29 30 31 32 33 34 35
	 */
	{
		float data[5][7];

		for (k = i = 0; i < 5; ++i) {
			for (j = 0; j < 7; ++j)
				data[i][j] = ++k;
		}

		/* write actual data to the dataset we just created */
		if (H5Dwrite(dsetid, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, data) < 0) errorexit();
	}

	/* define special values */
	{
		hid_t desc_tid, triplet_tid, sv_spaceid, sv_aid;
		hsize_t numspecialvalues = NUM_SPECIALVALUES;

		/* create a compound datatype for a special value */
		if ((triplet_tid = H5Tcreate(H5T_COMPOUND, sizeof(triplet_t))) < 0) errorexit();
		if ((desc_tid = H5Tcopy(H5T_C_S1)) < 0) errorexit();
		/* the description uses HDF5 variable length string. */
		if (H5Tset_size(desc_tid, H5T_VARIABLE) < 0) errorexit();
		/* insert description. */
		if (H5Tinsert(triplet_tid, "description", HOFFSET(triplet_t, description), desc_tid) < 0) errorexit();
		/* insert the special value */
		if (H5Tinsert(triplet_tid, "value", HOFFSET(triplet_t, value), H5T_NATIVE_INT) < 0) errorexit();
		/* insert the region reference. */
		if (H5Tinsert(triplet_tid, "region", HOFFSET(triplet_t, region), H5T_STD_REF_DSETREG) < 0) errorexit();

		/* create a dataspace for the special value attribute */
		if ((sv_spaceid = H5Screate_simple(1, &numspecialvalues, NULL)) < 0) errorexit();

		/* create the special value attribute */
		if ((sv_aid = H5Acreate(dsetid, "SpecialValues", triplet_tid, sv_spaceid, H5P_DEFAULT)) < 0) errorexit();

		/* write special values */
		{
			triplet_t buf[NUM_SPECIALVALUES];
			hsize_t start[DATASET_RANK], stride[DATASET_RANK], count[DATASET_RANK];

			/* set "water filled" */
			buf[0].description = "WaterFilled";
			buf[0].value = 1;

			/* mark the first, fourth and seventh columns
			        0  1  2  3  4  5  6 (X: marked)
			     +----------------------
			   0 |  X        X        X
			   1 |  X        X        X
			   2 |  X        X        X
			   3 |  X        X        X
			   4 |  X        X        X
			 */
			/* select the region for the first special value. */
			start[0] = 0; start[1] = 0;
			stride[0] = 1; stride[1] = 3;
			count[0] = 5; count[1] = 3;
			
			if (H5Sselect_hyperslab(spaceid, H5S_SELECT_SET, start, stride, count, NULL) < 0) errorexit();
			
			/* save the region with region reference. */
			if (H5Rcreate(&buf[0].region, fileid, "SnowDepth", H5R_DATASET_REGION, spaceid) < 0) errorexit();

			/* set "ice covered" */
			buf[1].description = "IceCovered";
			buf[1].value = 2;

			/* mark the first and sixth columns
			        0  1  2  3  4  5  6 (X: marked)
			     +----------------------
			   0 |  X              X
			   1 |  X              X
			   2 |  X              X
			   3 |  X              X
			   4 |  X              X
			 */
			/* select the region for the second special value.*/ 
			start[0] = 0; start[1] = 0;
			stride[0] = 1; stride[1] = 5;
			count[0] = 5; count[1] = 2;
			if (H5Sselect_hyperslab(spaceid, H5S_SELECT_SET, start, stride, count, NULL) < 0) errorexit();
			if (H5Rcreate(&buf[1].region, fileid, "SnowDepth", H5R_DATASET_REGION, spaceid) < 0) errorexit();

			/* write special values */
			if (H5Awrite(sv_aid, triplet_tid, buf) < 0) errorexit();
		}

		/* close the attribute */
		if (H5Aclose(sv_aid) < 0) errorexit();
		/* close the dataspace */
		if (H5Sclose(sv_spaceid) < 0) errorexit();
		/* close the datatype for a special value */
		if (H5Tclose(triplet_tid) < 0) errorexit();
		if (H5Tclose(desc_tid) < 0) errorexit();
	}

	/* close the dataset */
	if (H5Dclose(dsetid) < 0) errorexit();
	/* close the dataspace */
	if (H5Sclose(spaceid) < 0) errorexit();
	/* close the file */
	if (H5Fclose(fileid) < 0) errorexit();
}

int main(int argc, char **argv)
{
	create_file("regionsref.c.h5");
	return 0;
}

/* vi:set ts=4 sw=4 sts=4: */
